@@ -33,6 +33,7 @@
 #include <ufo/ufo-gpu-node.h>
 #include <ufo/ufo-remote-node.h>
 #include <ufo/ufo-enums.h>
+#include "ufo-priv.h"
 #include "compat.h"
 
 /**
@@ -41,8 +42,9 @@
  * @Title: UfoResources
  *
  * The #UfoResources creates the OpenCL environment and loads OpenCL kernels
- * from text files. Users should in general not create a resources object
- * themselves but use one that is created automatically by #UfoArchGraph.
+ * from disk or directly as a string. By default the kernel search path is in
+ * `$datadir/ufo` but can be extended by the `UFO_KERNEL_PATH` environment
+ * variable.
  */
 
 static void ufo_resources_initable_iface_init (GInitableIface *iface);
@@ -81,6 +83,7 @@ struct _UfoResourcesPrivate {
     cl_context       context;
     cl_uint          n_devices;         /* Number of OpenCL devices per platform id */
     cl_device_id     *devices;          /* Array of OpenCL devices per platform id */
+    gchar           **device_names;     /* Array of names for each device */
 
     GList       *gpu_nodes;
 
@@ -281,9 +284,13 @@ get_preferably_gpu_based_platform (UfoResourcesPrivate *priv)
     cl_uint n_platforms;
     cl_platform_id candidate = 0;
 
-    UFO_RESOURCES_CHECK_CLERR (clGetPlatformIDs (0, NULL, &n_platforms));
+    UFO_RESOURCES_CHECK_AND_SET (clGetPlatformIDs (0, NULL, &n_platforms), &priv->construct_error);
+
+    if (priv->construct_error != NULL)
+        return NULL;
+
     platforms = g_malloc0 (n_platforms * sizeof (cl_platform_id));
-    UFO_RESOURCES_CHECK_CLERR (clGetPlatformIDs (n_platforms, platforms, NULL));
+    UFO_RESOURCES_CHECK_AND_SET (clGetPlatformIDs (n_platforms, platforms, NULL), &priv->construct_error);
 
     g_debug ("INFO found %i OpenCL platforms %i", n_platforms, priv->platform_index);
 
@@ -326,17 +333,6 @@ platform_vendor_has_prefix (cl_platform_id platform,
     return has_prefix;
 }
 
-static void
-add_vendor_to_build_opts (GString *opts,
-                          cl_platform_id platform)
-{
-    if (platform_vendor_has_prefix (platform, "NVIDIA"))
-        g_string_append (opts, "-cl-nv-verbose -DVENDOR_NVIDIA");
-
-    if (platform_vendor_has_prefix (platform, "Advanced Micro Devices"))
-        g_string_append (opts, "-DVENDOR_AMD");
-}
-
 static void
 restrict_to_gpu_subset (UfoResourcesPrivate *priv)
 {
@@ -354,7 +350,7 @@ restrict_to_gpu_subset (UfoResourcesPrivate *priv)
 
     set = g_strsplit (var, ",", 0);
     n_set = g_strv_length (set);
-    device_indices = g_malloc0 (n_set);
+    device_indices = g_new0 (guint, n_set);
 
     n_devices = 0;
 
@@ -427,17 +423,32 @@ initialize_opencl (UfoResourcesPrivate *priv)
     cl_int errcode = CL_SUCCESS;
 
     priv->platform = get_preferably_gpu_based_platform (priv);
-    add_vendor_to_build_opts (priv->build_opts, priv->platform);
+
+    if (priv->platform == NULL)
+        return FALSE;
+
+    if (platform_vendor_has_prefix (priv->platform, "NVIDIA"))
+        g_string_append (priv->build_opts, "-cl-nv-verbose -DVENDOR_NVIDIA");
+
+    if (platform_vendor_has_prefix (priv->platform, "Advanced Micro Devices"))
+        g_string_append (priv->build_opts, "-DVENDOR_AMD");
 
     device_type = get_device_type_from_env ();
-    device_type |= priv->device_type & UFO_DEVICE_CPU ? CL_DEVICE_TYPE_CPU : 0;
-    device_type |= priv->device_type & UFO_DEVICE_GPU ? CL_DEVICE_TYPE_GPU : 0;
-    device_type |= priv->device_type & UFO_DEVICE_ACC ? CL_DEVICE_TYPE_ACCELERATOR : 0;
 
-    g_debug ("INFO Using CPUs=%i GPUs=%i Accelerators=%i",
-             (device_type & CL_DEVICE_TYPE_CPU) != 0,
-             (device_type & CL_DEVICE_TYPE_GPU) != 0,
-             (device_type & CL_DEVICE_TYPE_ACCELERATOR) != 0);
+    if (device_type == 0) {
+        /*
+         * If the user did not set anything from the outside, check the
+         * device-type property.
+         */
+        device_type |= priv->device_type & UFO_DEVICE_CPU ? CL_DEVICE_TYPE_CPU : 0;
+        device_type |= priv->device_type & UFO_DEVICE_GPU ? CL_DEVICE_TYPE_GPU : 0;
+        device_type |= priv->device_type & UFO_DEVICE_ACC ? CL_DEVICE_TYPE_ACCELERATOR : 0;
+    }
+
+    g_debug ("INFO Using CPUs=%c GPUs=%c Accelerators=%c",
+             (device_type & CL_DEVICE_TYPE_CPU) != 0 ? 'y' : 'n',
+             (device_type & CL_DEVICE_TYPE_GPU) != 0 ? 'y' : 'n',
+             (device_type & CL_DEVICE_TYPE_ACCELERATOR) != 0 ? 'y' : 'n');
 
     errcode = clGetDeviceIDs (priv->platform, device_type, 0, NULL, &priv->n_devices);
     UFO_RESOURCES_CHECK_AND_SET (errcode, &priv->construct_error);
@@ -455,17 +466,6 @@ initialize_opencl (UfoResourcesPrivate *priv)
 
     restrict_to_gpu_subset (priv);
 
-    for (guint i = 0; i < priv->n_devices; i++) {
-        size_t size;
-        gchar *name;
-
-        UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[i], CL_DEVICE_NAME, 0, NULL, &size));
-        name = g_malloc0 (size);
-
-        UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[i], CL_DEVICE_NAME, size, name, NULL));
-        g_free (name);
-    }
-
     priv->context = clCreateContext (NULL, priv->n_devices, priv->devices, NULL, NULL, &errcode);
     UFO_RESOURCES_CHECK_AND_SET (errcode, &priv->construct_error);
 
@@ -473,22 +473,21 @@ initialize_opencl (UfoResourcesPrivate *priv)
         return FALSE;
 
     priv->gpu_nodes = NULL;
+    priv->device_names = g_malloc0 (priv->n_devices * sizeof (gchar *));
 
     for (guint i = 0; i < priv->n_devices; i++) {
         UfoGpuNode *node;
         size_t size;
-        gchar *name;
 
         node = UFO_GPU_NODE (ufo_gpu_node_new (priv->context, priv->devices[i]));
 
         UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[i], CL_DEVICE_NAME, 0, NULL, &size));
-        name = g_malloc0 (size);
+        priv->device_names[i] = g_malloc0 (size);
 
-        UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[i], CL_DEVICE_NAME, size, name, NULL));
+        UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[i], CL_DEVICE_NAME, size, priv->device_names[i], NULL));
 
-        g_debug("NEW  UfoGpuNode-%p [device=%s]", (gpointer) node, name);
+        g_debug("NEW  UfoGpuNode-%p [device=%s]", (gpointer) node, priv->device_names[i]);
         priv->gpu_nodes = g_list_append (priv->gpu_nodes, node);
-        g_free (name);
     }
 
     return TRUE;
@@ -508,6 +507,11 @@ ufo_resources_new (GError **error)
     return g_initable_new (UFO_TYPE_RESOURCES, NULL, error, NULL);
 }
 
+/**
+ * ufo_resources_add_path:
+ * @resources: A #UfoResources object
+ * @path: a path to OpenCL kernel files
+ */
 void
 ufo_resources_add_path (UfoResources *resources,
                         const gchar *path)
@@ -518,25 +522,6 @@ ufo_resources_add_path (UfoResources *resources,
     resources->priv->paths = g_list_append (resources->priv->paths, g_strdup (path));
 }
 
-static gchar *
-escape_device_name (gchar *name)
-{
-    gchar *tmp = name;
-
-    while (*tmp) {
-        gchar c = *tmp;
-
-        if (c == ' ' || c == '(' || c == ')' || c == '-' || c == '@' || c == '.')
-            *tmp = '_';
-        else
-            *tmp = g_ascii_toupper (c);
-
-        tmp++;
-    }
-
-    return name;
-}
-
 static void
 append_include_path (const gchar *path,
                      GString *str)
@@ -544,34 +529,21 @@ append_include_path (const gchar *path,
     g_string_append_printf (str, " -I%s", path);
 }
 
-static gchar *
-get_device_build_options (UfoResourcesPrivate *priv,
-                          guint device_index,
-                          const gchar *additional)
+static void
+opt_append_include_paths (GString *str,
+                          UfoResourcesPrivate *priv)
+
 {
-    GString *opts;
-    gsize size;
-    gchar *name;
+    g_list_foreach (priv->paths, (GFunc) append_include_path, str);
+}
 
+static void
+opt_append_device_options (GString *str,
+                           UfoResourcesPrivate *priv,
+                           guint device_index)
+{
     g_assert (device_index < priv->n_devices);
-
-    opts = g_string_new (priv->build_opts->str);
-
-    if (additional != NULL && strlen (additional) > 0) {
-        g_string_append_printf (opts, " %s", additional);
-    }
-
-    UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[device_index], CL_DEVICE_NAME, 0, NULL, &size));
-    name = g_malloc0 (size);
-
-    UFO_RESOURCES_CHECK_CLERR (clGetDeviceInfo (priv->devices[device_index], CL_DEVICE_NAME, size, name, NULL));
-
-    g_string_append_printf (opts, " -DDEVICE_%s", escape_device_name (name));
-    g_free (name);
-
-    g_list_foreach (priv->paths, (GFunc) append_include_path, opts);
-
-    return g_string_free (opts, FALSE);
+    g_string_append_printf (str, " -DDEVICE_%s", ufo_escape_device_name (priv->device_names[device_index]));
 }
 
 static void
@@ -603,7 +575,6 @@ add_program_from_source (UfoResourcesPrivate *priv,
 {
     cl_program program;
     cl_int errcode = CL_SUCCESS;
-    gchar *build_options;
     GTimer *timer;
 
     program = g_hash_table_lookup (priv->programs, source);
@@ -619,26 +590,42 @@ add_program_from_source (UfoResourcesPrivate *priv,
         return NULL;
     }
 
-    build_options = get_device_build_options (priv, 0, options);
     timer = g_timer_new ();
 
-    errcode = clBuildProgram (program,
-                              priv->n_devices, priv->devices,
-                              build_options,
-                              NULL, NULL);
+    for (guint i = 0; i < priv->n_devices; i++) {
+        GString *build_options;
 
-    g_timer_stop (timer);
-    g_debug ("INFO Built with `%s' in %3.5fs", build_options, g_timer_elapsed (timer, NULL));
-    g_timer_destroy (timer);
+        build_options = g_string_new (priv->build_opts->str);
+        opt_append_device_options (build_options, priv, i);
+        opt_append_include_paths (build_options, priv);
 
-    if (errcode != CL_SUCCESS) {
-        handle_build_error (program, priv->devices[0], errcode, error);
-        return NULL;
+        if (options) {
+            gchar *stripped_opts = g_strstrip (g_strdup (options));
+            g_string_append (build_options, " ");
+            g_string_append (build_options, stripped_opts);
+            g_free (stripped_opts);
+        }
+
+        errcode = clBuildProgram (program, 1, &priv->devices[i], build_options->str, NULL, NULL);
+
+        if (errcode != CL_SUCCESS) {
+            handle_build_error (program, priv->devices[0], errcode, error);
+            return NULL;
+        }
+
+        if (options)
+            g_debug ("INFO Built with `%s %s' for device %i", build_options->str, options, i);
+        else
+            g_debug ("INFO Built with `%s' for device %i", build_options->str, i);
+
+        g_string_free (build_options, TRUE);
     }
 
+    g_timer_stop (timer);
+    g_debug ("INFO Built program for %i devices in %3.5fs", priv->n_devices, g_timer_elapsed (timer, NULL));
+    g_timer_destroy (timer);
     g_hash_table_insert (priv->programs, g_strdup (source), program);
 
-    g_free (build_options);
     return program;
 }
 
@@ -720,10 +707,12 @@ create_cache_key (const gchar *filename,
  * @error: Return location for a GError from #UfoResourcesError, or %NULL
  *
  * Loads a and builds a kernel from a file. The file is searched in the current
- * working directory and all paths added through ufo_resources_add_paths (). If
+ * working directory and all paths added through ufo_resources_add_paths () as
+ * well as paths specified in the `UFO_KERNEL_PATH` environment variable. If
  * @kernel is %NULL, the first encountered kernel is returned.
  *
- * Returns: (transfer none): a cl_kernel object that is load from @filename or %NULL on error
+ * Returns: (transfer none): a cl_kernel object that is load from @filename or
+ *  %NULL on error
  */
 gpointer
 ufo_resources_get_kernel_with_opts (UfoResources   *resources,
@@ -785,7 +774,8 @@ exit:
  * working directory and all paths added through ufo_resources_add_path (). If
  * @kernel is %NULL, the first encountered kernel is returned.
  *
- * Returns: (transfer none): a cl_kernel object that is load from @filename or %NULL on error
+ * Returns: (transfer none): a cl_kernel object that is load from @filename or
+ *  %NULL on error
  */
 gpointer
 ufo_resources_get_kernel (UfoResources *resources,
@@ -808,7 +798,8 @@ ufo_resources_get_kernel (UfoResources *resources,
  * @kernel is %NULL, the first encountered kernel is returned. The kernel object
  * is cached and should not be used by two threads concurrently.
  *
- * Returns: (transfer none): a cl_kernel object that is load from @filename or %NULL on error
+ * Returns: (transfer none): a cl_kernel object that is load from @filename or
+ *  %NULL on error
  */
 gpointer
 ufo_resources_get_cached_kernel (UfoResources *resources,
@@ -848,6 +839,42 @@ ufo_resources_get_cached_kernel (UfoResources *resources,
     return kernel;
 }
 
+/**
+ * ufo_resources_get_kernel_from_source_with_opts:
+ * @resources: A #UfoResources
+ * @source: OpenCL source string
+ * @kernel: Name of a kernel or %NULL
+ * @options: Options passed to the OpenCL compiler
+ * @error: Return location for a GError from #UfoResourcesError, or NULL
+ *
+ * Loads and builds a kernel from a string. If @kernel is %NULL, the first
+ * kernel defined in @source is used.
+ *
+ * Returns: (transfer none): a cl_kernel object that is load from @filename
+ */
+gpointer
+ufo_resources_get_kernel_from_source_with_opts (UfoResources *resources,
+                                                const gchar *source,
+                                                const gchar *kernel,
+                                                const gchar *options,
+                                                GError **error)
+{
+    UfoResourcesPrivate *priv;
+    cl_program program;
+
+    g_return_val_if_fail (UFO_IS_RESOURCES (resources) &&
+                          (source != NULL), NULL);
+
+    priv = UFO_RESOURCES_GET_PRIVATE (resources);
+    program = add_program_from_source (priv, source, options, error);
+
+    if (program == NULL)
+        return NULL;
+
+    g_debug ("INFO Added program %p from source", (gpointer) program);
+    return create_kernel (priv, program, kernel, error);
+}
+
 /**
  * ufo_resources_get_kernel_from_source:
  * @resources: A #UfoResources
@@ -866,20 +893,7 @@ ufo_resources_get_kernel_from_source (UfoResources *resources,
                                       const gchar *kernel,
                                       GError **error)
 {
-    UfoResourcesPrivate *priv;
-    cl_program program;
-
-    g_return_val_if_fail (UFO_IS_RESOURCES (resources) &&
-                          (source != NULL), NULL);
-
-    priv = UFO_RESOURCES_GET_PRIVATE (resources);
-    program = add_program_from_source (priv, source, NULL, error);
-
-    if (program == NULL)
-        return NULL;
-
-    g_debug ("INFO Added program %p from source", (gpointer) program);
-    return create_kernel (priv, program, kernel, error);
+    return ufo_resources_get_kernel_from_source_with_opts (resources, source, kernel, NULL, error);
 }
 
 /**
@@ -903,8 +917,7 @@ ufo_resources_get_kernel_source (UfoResources   *resources,
     gchar *path;
     gchar *buffer;
 
-    g_return_val_if_fail (UFO_IS_RESOURCES (resources) &&
-                          (filename != NULL), NULL);
+    g_return_val_if_fail (UFO_IS_RESOURCES (resources) && (filename != NULL), NULL);
 
     buffer = NULL;
     priv = resources->priv;
@@ -921,10 +934,8 @@ ufo_resources_get_kernel_source (UfoResources   *resources,
     if (buffer == NULL) {
         g_set_error (error, UFO_RESOURCES_ERROR, UFO_RESOURCES_ERROR_LOAD_PROGRAM,
                      "Could not open `%s'", filename);
-        goto exit;
     }
 
-exit:
     g_free (path);
     return buffer;
 }
@@ -1140,6 +1151,11 @@ ufo_resources_finalize (GObject *object)
 
     g_hash_table_destroy (priv->programs);
 
+    if (priv->device_names != NULL) {
+        for (guint i = 0; i < priv->n_devices; i++)
+            g_free (priv->device_names[i]);
+    }
+
     if (priv->context) {
         g_debug ("FREE context=%p", (gpointer) priv->context);
         UFO_RESOURCES_CHECK_CLERR (clReleaseContext (priv->context));
@@ -1147,6 +1163,7 @@ ufo_resources_finalize (GObject *object)
 
     g_string_free (priv->build_opts, TRUE);
 
+    g_free (priv->device_names);
     g_free (priv->devices);
 
     priv->kernels = NULL;
@@ -1174,8 +1191,15 @@ ufo_resources_initable_init (GInitable *initable,
     resources = UFO_RESOURCES (initable);
     priv = resources->priv;
 
-    if (priv->construct_error != NULL)
+    if (priv->construct_error != NULL) {
+        /* 
+         * For some very strange reason, using g_propagate_error() does not
+         * work, the error variable is set but does not contain the message.
+         */
+        g_set_error_literal (error, UFO_RESOURCES_ERROR, UFO_RESOURCES_ERROR_GENERAL,
+                             priv->construct_error->message);
         return FALSE;
+    }
 
     return TRUE;
 }
