@@ -62,6 +62,7 @@ typedef struct {
     guint           *dims;
     gboolean        *finished;
     gboolean         strict;
+    gboolean         timestamps;
 } TaskLocalData;
 
 
@@ -299,7 +300,18 @@ run_task (TaskLocalData *tld)
                 break;
 
             case UFO_TASK_MODE_GENERATOR:
-                active = ufo_task_generate (tld->task, output, &requisition);
+                {
+                    if (tld->timestamps) {
+                        GValue v = { 0, };
+
+                        g_value_init (&v, G_TYPE_INT64);
+                        g_value_set_int64 (&v, g_get_real_time ());
+                        ufo_buffer_set_metadata (output, "ts", &v);
+                    }
+
+                    active = ufo_task_generate (tld->task, output, &requisition);
+
+                }
                 break;
 
             default:
@@ -391,10 +403,18 @@ setup_tasks (UfoBaseScheduler *scheduler,
     TaskLocalData **tlds;
     GList *nodes;
     guint n_nodes;
+    gboolean timestamps;
     gboolean tracing_enabled;
 
-    resources = ufo_base_scheduler_get_resources (scheduler);
-    g_object_get (scheduler, "enable-tracing", &tracing_enabled, NULL);
+    resources = ufo_base_scheduler_get_resources (scheduler, error);
+
+    if (resources == NULL)
+        return NULL;
+
+    g_object_get (scheduler,
+                  "enable-tracing", &tracing_enabled,
+                  "timestamps", &timestamps,
+                  NULL);
 
     nodes = ufo_graph_get_nodes (UFO_GRAPH (task_graph));
     n_nodes = g_list_length (nodes);
@@ -414,6 +434,7 @@ setup_tasks (UfoBaseScheduler *scheduler,
         tld->mode = ufo_task_get_mode (tld->task);
         tld->n_inputs = ufo_task_get_num_inputs (tld->task);
         tld->dims = g_new0 (guint, tld->n_inputs);
+        tld->timestamps = timestamps;
 
         /* TODO: make this configurable from outside */
         tld->strict = FALSE;
@@ -439,7 +460,8 @@ setup_tasks (UfoBaseScheduler *scheduler,
 
 static GList *
 setup_groups (UfoBaseScheduler *scheduler,
-              UfoTaskGraph *task_graph)
+              UfoTaskGraph *task_graph,
+              GError **error)
 {
     UfoResources *resources;
     GList *groups;
@@ -449,7 +471,11 @@ setup_groups (UfoBaseScheduler *scheduler,
 
     groups = NULL;
     nodes = ufo_graph_get_nodes (UFO_GRAPH (task_graph));
-    resources = ufo_base_scheduler_get_resources (scheduler);
+    resources = ufo_base_scheduler_get_resources (scheduler, error);
+
+    if (resources == NULL)
+        return NULL;
+
     context = ufo_resources_get_context (resources);
 
     g_list_for (nodes, it) {
@@ -589,22 +615,21 @@ ufo_scheduler_run (UfoBaseScheduler *scheduler,
     GThread **threads;
     TaskLocalData **tlds;
     gboolean expand;
-    gboolean trace;
 
     priv = UFO_SCHEDULER_GET_PRIVATE (scheduler);
 
-    g_object_get (scheduler,
-                  "enable-tracing", &trace,
-                  "expand", &expand,
-                  NULL);
+    g_object_get (scheduler, "expand", &expand, NULL);
 
     graph = task_graph;
-    resources = ufo_base_scheduler_get_resources (scheduler);
+    resources = ufo_base_scheduler_get_resources (scheduler, error);
+
+    if (resources == NULL)
+        return;
+
     gpu_nodes = ufo_resources_get_gpu_nodes (resources);
 
-    if (priv->mode == UFO_REMOTE_MODE_REPLICATE) {
+    if (priv->mode == UFO_REMOTE_MODE_REPLICATE)
         replicate_task_graph (graph, resources);
-    }
 
     if (expand) {
         gboolean expand_remote = priv->mode == UFO_REMOTE_MODE_STREAM;
@@ -621,11 +646,13 @@ ufo_scheduler_run (UfoBaseScheduler *scheduler,
     /* Prepare task structures */
     tlds = setup_tasks (scheduler, graph, error);
 
-    if (tlds == NULL) {
+    if (tlds == NULL)
         return;
-    }
 
-    groups = setup_groups (scheduler, graph);
+    groups = setup_groups (scheduler, graph, error);
+
+    if (groups == NULL)
+        return;
 
     if (!correct_connections (graph, error))
         return;
