@@ -32,6 +32,10 @@
  * SECTION:ufo-task-graph
  * @Short_description: Hold task nodes
  * @Title: UfoTaskGraph
+ *
+ * The task graph is the central data structure that connects #UfoTaskNode
+ * objects to form computational pipelines and graphs. To execute a task graph,
+ * it has to be passed to a #UfoBaseScheduler.
  */
 
 G_DEFINE_TYPE (UfoTaskGraph, ufo_task_graph, UFO_TYPE_GRAPH)
@@ -40,7 +44,6 @@ G_DEFINE_TYPE (UfoTaskGraph, ufo_task_graph, UFO_TYPE_GRAPH)
 
 struct _UfoTaskGraphPrivate {
     UfoPluginManager *manager;
-    GHashTable *prop_sets;
     GHashTable *json_nodes;
     GList *remote_tasks;
     guint index;
@@ -52,20 +55,25 @@ typedef enum {
     JSON_DATA
 } JsonLocation;
 
+typedef struct {
+    UfoPluginManager *manager;
+    UfoTaskNode *task;
+    GError **error;
+} JsonSetPropertyInfo;
+
 static void add_nodes_from_json     (UfoTaskGraph *self, JsonNode *, GError **);
-static void handle_json_prop_set    (JsonObject *, const gchar *, JsonNode *, gpointer user);
 static void handle_json_task_edge   (JsonArray *, guint, JsonNode *, gpointer);
 static void add_task_node_to_json_array (UfoTaskNode *, JsonArray *);
 static JsonObject *json_object_from_ufo_node (UfoNode *node);
 static JsonNode *get_json_representation (UfoTaskGraph *, GError **);
-static UfoTaskNode *create_node_from_json (JsonNode *json_node,UfoPluginManager *manager, GHashTable *prop_sets, GError **error);
-static void add_node_to_table (JsonObject *object, const gchar *name, JsonNode *node, gpointer table);
+static UfoTaskNode *create_node_from_json (JsonNode *json_node,UfoPluginManager *manager, GError **error);
 
 /*
  * ChangeLog:
  * - 1.1: Add "index" and "total" keys to the root object
+ * - 2.0: Add "index" and "total" keys to the root object
  */
-static const gchar *JSON_API_VERSION = "1.1";
+static const gchar *JSON_API_VERSION = "2.0";
 
 /**
  * UfoTaskGraphError:
@@ -112,16 +120,11 @@ read_json (UfoTaskGraph *graph,
 
     switch (location) {
         case JSON_FILE:
-            json_parser_load_from_file (json_parser,
-                                        data,
-                                        &tmp_error);
+            json_parser_load_from_file (json_parser, data, &tmp_error);
             break;
 
         case JSON_DATA:
-            json_parser_load_from_data (json_parser,
-                                        data,
-                                        (gssize) strlen (data),
-                                        &tmp_error);
+            json_parser_load_from_data (json_parser, data, (gssize) strlen (data), &tmp_error);
             break;
     }
 
@@ -163,10 +166,7 @@ ufo_task_graph_read_from_file (UfoTaskGraph *graph,
                                const gchar *filename,
                                GError **error)
 {
-    g_return_if_fail (UFO_IS_TASK_GRAPH (graph) &&
-                      UFO_IS_PLUGIN_MANAGER (manager) &&
-                      (filename != NULL));
-
+    g_return_if_fail (UFO_IS_TASK_GRAPH (graph) && UFO_IS_PLUGIN_MANAGER (manager) && (filename != NULL));
     read_json (graph, manager, JSON_FILE, filename, error);
 }
 
@@ -185,10 +185,7 @@ ufo_task_graph_read_from_data (UfoTaskGraph *graph,
                                const gchar *json,
                                GError **error)
 {
-    g_return_if_fail (UFO_IS_TASK_GRAPH (graph) &&
-                      UFO_IS_PLUGIN_MANAGER (manager) &&
-                      (json != NULL));
-
+    g_return_if_fail (UFO_IS_TASK_GRAPH (graph) && UFO_IS_PLUGIN_MANAGER (manager) && (json != NULL));
     read_json (graph, manager, JSON_DATA, json, error);
 }
 
@@ -269,7 +266,7 @@ task_graph_to_generator (UfoTaskGraph *graph,
 
 /**
  * ufo_task_graph_save_to_json:
- * @graph: A #UfoTaskGraph.
+ * @graph: A #UfoTaskGraph
  * @filename: Path and filename to the JSON file
  * @error: Indicates error in case of failed file saving
  *
@@ -290,6 +287,15 @@ ufo_task_graph_save_to_json (UfoTaskGraph *graph,
     }
 }
 
+/**
+ * ufo_task_graph_get_json_data:
+ * @graph: A #UfoTaskGraph
+ * @error: Indicates error in case of failed serialization
+ *
+ * Serialize @graph to a JSON string.
+ *
+ * Returns: (transfer full): A JSON string describing @graph
+ */
 gchar *
 ufo_task_graph_get_json_data (UfoTaskGraph *graph,
                               GError **error)
@@ -392,78 +398,83 @@ expand_remotes (UfoTaskGraph *task_graph,
     g_object_unref (remote_graph);
 }
 
-static gboolean
-has_common_ancestries (UfoTaskGraph *graph, GList *path)
+static GList *
+nodes_with_common_ancestries (UfoTaskGraph *graph, GList *path)
 {
+    GList *result = NULL;
     GList *it;
 
     g_list_for (path, it) {
         if (ufo_graph_get_num_predecessors (UFO_GRAPH (graph), UFO_NODE (it->data)) > 1)
-            return TRUE;
+            result = g_list_append (result, it->data);
     }
 
-    return FALSE;
+    return result;
 }
 
 /**
  * ufo_task_graph_expand:
- * @task_graph: A #UfoTaskGraph
+ * @graph: A #UfoTaskGraph
  * @resources: A #UfoResources objects
  * @n_gpus: Number of GPUs to expand the graph for
  * @expand_remote: %TRUE if remote nodes should be inserted
  *
- * Expands @task_graph in a way that most of the resources in @arch_graph can be
- * occupied. In the simple pipeline case, the longest possible GPU paths are
- * duplicated as much as there are GPUs in @arch_graph.
+ * Expands @graph in a way that most of the resources in @graph can be occupied.
+ * In the simple pipeline case, the longest possible GPU paths are duplicated as
+ * much as there are GPUs in @arch_graph.
  */
 void
-ufo_task_graph_expand (UfoTaskGraph *task_graph,
+ufo_task_graph_expand (UfoTaskGraph *graph,
                        UfoResources *resources,
                        guint n_gpus,
                        gboolean expand_remote)
 {
     GList *path;
-    GList *roots;
-    guint num_roots;
+    GList *common;
 
-    g_return_if_fail (UFO_IS_TASK_GRAPH (task_graph));
+    g_return_if_fail (UFO_IS_TASK_GRAPH (graph));
 
-    /*
-     * Check if we start with multiple roots. If so, do not expand and map from
-     * the beginning.
-     */
-    roots = ufo_graph_get_roots (UFO_GRAPH (task_graph));
-    num_roots = g_list_length (roots);
-    g_list_free (roots);
+    path = ufo_graph_find_longest_path (UFO_GRAPH (graph), (UfoFilterPredicate) is_gpu_task, NULL);
+    common = nodes_with_common_ancestries (graph, path);
 
-    if (num_roots > 1)
-        return;
-
-    path = ufo_graph_find_longest_path (UFO_GRAPH (task_graph), (UfoFilterPredicate) is_gpu_task, NULL);
-
-    /*
-     * Check if any node on the path contains multiple inputs and stop expansion
-     * TODO: we need a better strategy at this point ...
-     */
-    if (has_common_ancestries (task_graph, path)) {
+    if (g_list_length (common) > 1) {
+        g_debug ("WARN More than one nodes have multiple inputs, not going to expand");
+        g_list_free (common);
         g_list_free (path);
         return;
     }
 
+    if (g_list_length (common) == 1) {
+        GList *it;
+
+        g_debug ("INFO Found node with multiple inputs, going to prune it");
+        it = g_list_first (path);
+
+        while (it->data != common->data) {
+            GList *jt = g_list_next (it);
+
+            path = g_list_remove_link (path, it);
+            it = jt;
+        }
+
+        /* remove the common node as well */
+        g_object_unref (it->data);
+        path = g_list_remove_link (path, it);
+        g_list_free (common);
+    }
+
     if (path != NULL && g_list_length (path) > 0) {
         GList *predecessors;
         GList *successors;
 
-        g_object_unref (UFO_NODE (g_list_first (path)->data));
-
         if (g_list_length (path) > 1)
             g_object_unref (UFO_NODE (g_list_last (path)->data));
 
         /* Add predecessor and successor nodes to path */
-        predecessors = ufo_graph_get_predecessors (UFO_GRAPH (task_graph),
+        predecessors = ufo_graph_get_predecessors (UFO_GRAPH (graph),
                                                    UFO_NODE (g_list_first (path)->data));
 
-        successors = ufo_graph_get_successors (UFO_GRAPH (task_graph),
+        successors = ufo_graph_get_successors (UFO_GRAPH (graph),
                                                UFO_NODE (g_list_last (path)->data));
 
         if (predecessors != NULL)
@@ -484,7 +495,7 @@ ufo_task_graph_expand (UfoTaskGraph *task_graph,
 
             if (n_remotes > 0) {
                 g_debug ("INFO Expand for %i remote nodes", n_remotes);
-                expand_remotes (task_graph, remotes, path);
+                expand_remotes (graph, remotes, path);
             }
 
             g_list_free (remotes);
@@ -493,7 +504,7 @@ ufo_task_graph_expand (UfoTaskGraph *task_graph,
         g_debug ("INFO Expand for %i GPU nodes", n_gpus);
 
         for (guint i = 1; i < n_gpus; i++)
-            ufo_graph_expand (UFO_GRAPH (task_graph), path);
+            ufo_graph_expand (UFO_GRAPH (graph), path);
     }
 
     g_list_free (path);
@@ -501,14 +512,14 @@ ufo_task_graph_expand (UfoTaskGraph *task_graph,
 
 /**
  * ufo_task_graph_fuse:
- * @task_graph: A #UfoTaskGraph
+ * @graph: A #UfoTaskGraph
  *
  * Fuses task nodes to increase data locality.
  *
  * Note: This is not implemented and a no-op right now.
  */
 void
-ufo_task_graph_fuse (UfoTaskGraph *task_graph)
+ufo_task_graph_fuse (UfoTaskGraph *graph)
 {
 }
 
@@ -549,7 +560,7 @@ map_proc_node (UfoGraph *graph,
 
 /**
  * ufo_task_graph_is_alright:
- * @task_graph: A #UfoTaskGraph
+ * @graph: A #UfoTaskGraph
  * @error: Location for a GError or %NULL
  *
  * Check if nodes int the task graph are properly connected.
@@ -557,20 +568,20 @@ map_proc_node (UfoGraph *graph,
  * Returns: %TRUE if everything is alright, %FALSE else.
  */
 gboolean
-ufo_task_graph_is_alright (UfoTaskGraph *task_graph,
+ufo_task_graph_is_alright (UfoTaskGraph *graph,
                            GError **error)
 {
     GList *nodes;
     GList *it;
     gboolean alright = TRUE;
 
-    nodes = ufo_graph_get_nodes (UFO_GRAPH (task_graph));
+    nodes = ufo_graph_get_nodes (UFO_GRAPH (graph));
 
     /* Check that nodes don't receive input from reductors and processors */
     g_list_for (nodes, it) {
         GList *predecessors;
 
-        predecessors = ufo_graph_get_predecessors (UFO_GRAPH (task_graph), UFO_NODE (it->data));
+        predecessors = ufo_graph_get_predecessors (UFO_GRAPH (graph), UFO_NODE (it->data));
 
         if (g_list_length (predecessors) > 1) {
             GList *jt;
@@ -580,16 +591,8 @@ ufo_task_graph_is_alright (UfoTaskGraph *task_graph,
                 combined_modes |= ufo_task_get_mode (UFO_TASK (jt->data));
 
             if ((combined_modes & UFO_TASK_MODE_PROCESSOR) && (combined_modes & UFO_TASK_MODE_REDUCTOR)) {
-#if 0
-                g_set_error (error, UFO_TASK_GRAPH_ERROR, UFO_TASK_GRAPH_ERROR_BAD_INPUTS,
-                             "`%s' receives both processor and reductor inputs which may deadlock.",
-                             ufo_task_node_get_plugin_name (UFO_TASK_NODE (it->data)));
-                g_list_free (predecessors);
-                g_list_free (nodes);
-                return FALSE;
-#endif
-                g_warning ("`%s' receives both processor and reductor inputs which may deadlock.",
-                           ufo_task_node_get_plugin_name (UFO_TASK_NODE (it->data)));
+                g_debug ("`%s' receives both processor and reductor inputs which may deadlock.",
+                         ufo_task_node_get_plugin_name (UFO_TASK_NODE (it->data)));
             }
         }
 
@@ -599,7 +602,7 @@ ufo_task_graph_is_alright (UfoTaskGraph *task_graph,
     g_list_free (nodes);
 
     /* Check leaves are sinks */
-    nodes = ufo_graph_get_leaves (UFO_GRAPH (task_graph));
+    nodes = ufo_graph_get_leaves (UFO_GRAPH (graph));
 
     g_list_for (nodes, it) {
         if ((ufo_task_get_mode (UFO_TASK (it->data)) & UFO_TASK_MODE_TYPE_MASK) != UFO_TASK_MODE_SINK) {
@@ -618,23 +621,22 @@ ufo_task_graph_is_alright (UfoTaskGraph *task_graph,
 
 /**
  * ufo_task_graph_map:
- * @task_graph: A #UfoTaskGraph
+ * @graph: A #UfoTaskGraph
  * @gpu_nodes: (transfer none) (element-type Ufo.GpuNode): List of #UfoGpuNode objects
  *
- * Map task nodes of @task_graph to the processing nodes of @arch_graph. Not
- * doing this could break execution of @task_graph.
+ * Map task nodes of @graph to the list of @gpu_nodes.
  */
 void
-ufo_task_graph_map (UfoTaskGraph *task_graph,
+ufo_task_graph_map (UfoTaskGraph *graph,
                     GList *gpu_nodes)
 {
     GList *roots;
     GList *it;
 
-    roots = ufo_graph_get_roots (UFO_GRAPH (task_graph));
+    roots = ufo_graph_get_roots (UFO_GRAPH (graph));
 
     g_list_for (roots, it) {
-        map_proc_node (UFO_GRAPH (task_graph), UFO_NODE (it->data), 0, gpu_nodes);
+        map_proc_node (UFO_GRAPH (graph), UFO_NODE (it->data), 0, gpu_nodes);
     }
 
     g_list_free (roots);
@@ -676,6 +678,14 @@ ufo_task_graph_connect_nodes_full (UfoTaskGraph *graph,
     ufo_graph_connect_nodes (UFO_GRAPH (graph), UFO_NODE (n1), UFO_NODE (n2), GINT_TO_POINTER (input));
 }
 
+/**
+ * ufo_task_graph_set_partition:
+ * @graph: A #UfoTaskGraph
+ * @index: index of @total partitions - 1
+ * @total: total number of partitions
+ *
+ * Set the partition of this task graph.
+ */
 void
 ufo_task_graph_set_partition (UfoTaskGraph *graph,
                               guint index,
@@ -687,6 +697,14 @@ ufo_task_graph_set_partition (UfoTaskGraph *graph,
     graph->priv->total = total;
 }
 
+/**
+ * ufo_task_graph_get_partition:
+ * @graph: A #UfoTaskGraph
+ * @index: Location to store index
+ * @total: Location to store the total number of partitions
+ *
+ * Get the partition structure of @graph.
+ */
 void
 ufo_task_graph_get_partition (UfoTaskGraph *graph,
                               guint *index,
@@ -705,23 +723,22 @@ add_nodes_from_json (UfoTaskGraph *self,
     UfoTaskGraphPrivate *priv = UFO_TASK_GRAPH_GET_PRIVATE(self);
     JsonObject *root_object = json_node_get_object (root);
 
-    if (json_object_has_member (root_object, "prop-sets")) {
-        JsonObject *sets = json_object_get_object_member (root_object, "prop-sets");
-        json_object_foreach_member (sets, handle_json_prop_set, priv);
-    }
-
     if (json_object_has_member (root_object, "nodes")) {
         JsonArray *nodes = json_object_get_array_member (root_object, "nodes");
         GList *elements = json_array_get_elements (nodes);
         GList *it;
 
         g_list_for (elements, it) {
-            UfoTaskNode *new_node = create_node_from_json (it->data, priv->manager, priv->prop_sets, error);
-            if(new_node != NULL) {
-                const char *name = ufo_task_node_get_identifier(new_node);
+            UfoTaskNode *new_node = create_node_from_json (it->data, priv->manager, error);
+
+            if (new_node != NULL) {
+                const char *name = ufo_task_node_get_identifier (new_node);
 
                 if (g_hash_table_lookup (priv->json_nodes, name) != NULL) {
-                    g_error ("Duplicate name `%s' found", name);
+                    g_object_unref (new_node);
+                    g_set_error (error, UFO_TASK_GRAPH_ERROR, UFO_TASK_GRAPH_ERROR_JSON_KEY,
+                                 "Duplicate name `%s' found", name);
+                    return;
                 }
 
                 g_hash_table_insert (priv->json_nodes, g_strdup (name), new_node);
@@ -744,24 +761,107 @@ add_nodes_from_json (UfoTaskGraph *self,
     }
 }
 
+static void
+set_property_from_json (JsonObject *object, const gchar *name, JsonNode *node, JsonSetPropertyInfo *info)
+{
+    GParamSpec *pspec;
+
+    /* If error is already set before we just return and skip this node. */
+    if (info->error && *info->error)
+        return;
+
+    pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (info->task), name);
+
+    if (pspec == NULL) {
+        g_set_error (info->error, UFO_TASK_GRAPH_ERROR, UFO_TASK_GRAPH_ERROR_JSON_KEY,
+                     "Property `%s' does not exist", name);
+        return;
+    }
+
+    if (JSON_NODE_HOLDS_VALUE (node)) {
+        GValue val = {0,};
+
+        json_node_get_value (node, &val);
+        g_object_set_property (G_OBJECT (info->task), name, &val);
+        g_value_unset (&val);
+    }
+    else if (JSON_NODE_HOLDS_ARRAY (node)) {
+        GParamSpecValueArray *vapspec;
+        JsonArray *json_array;
+        GValueArray *value_array;
+        GList *values;
+        GList *it;
+        GValue value_array_val = {0, };
+
+        pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (info->task), name);
+
+        if (pspec->value_type != G_TYPE_VALUE_ARRAY) {
+            g_set_error (info->error, UFO_TASK_GRAPH_ERROR, UFO_TASK_GRAPH_ERROR_JSON_KEY,
+                         "`%s' is not an array", name);
+            return;
+        }
+
+        vapspec = (GParamSpecValueArray *) pspec;
+        json_array = json_node_get_array (node);
+        values = json_array_get_elements (json_array);
+        value_array = g_value_array_new (g_list_length (values));
+
+        g_list_for (values, it) {
+            GValue val = {0, };
+            GValue target_val = {0, };
+
+            json_node_get_value ((JsonNode *) it->data, &val);
+
+            /*
+             * We explicitly transform the value unconditionally to the
+             * target type because the type system is pretty strict and it
+             * is not possible to set a float property with a double value.
+             */
+            g_value_init (&target_val, vapspec->element_spec->value_type);
+            g_value_transform (&val, &target_val);
+            g_value_array_append (value_array, &target_val);
+            g_value_unset (&target_val);
+            g_value_unset (&val);
+        }
+
+        g_value_init (&value_array_val, G_TYPE_VALUE_ARRAY);
+        g_value_set_boxed (&value_array_val, value_array);
+        g_object_set_property (G_OBJECT (info->task), name, &value_array_val);
+        g_value_unset (&value_array_val);
+        g_list_free (values);
+        g_value_array_free (value_array);
+    }
+    else if (JSON_NODE_HOLDS_OBJECT (node)) {
+        JsonObject *node_object = json_node_get_object (node);
+
+        if (json_object_has_member (node_object, "plugin")) {
+            UfoTaskNode *inner_node = create_node_from_json (node, info->manager, info->error);
+            g_object_force_floating (G_OBJECT (inner_node));
+            g_object_set (G_OBJECT (info->task), name, inner_node, NULL);
+        }
+        else {
+            ufo_task_set_json_object_property (UFO_TASK (node), name, node_object);
+        }
+    }
+    else {
+        g_warning ("`%s' is neither a primitive value, array or object", name);
+    }
+}
+
 static UfoTaskNode *
 create_node_from_json (JsonNode *json_node,
                        UfoPluginManager *manager,
-                       GHashTable *prop_sets,
                        GError **error)
 {
     UfoTaskNode *ret_node = NULL;
     JsonObject *json_object;
     GError *tmp_error = NULL;
-    GHashTableIter iter;
-    gpointer key, value;
     const gchar *plugin_name;
     const gchar *task_name;
 
     json_object = json_node_get_object (json_node);
 
-    if (!json_object_has_member (json_object, "plugin") ||
-        !json_object_has_member (json_object, "name")) {
+    if (!json_object_has_member (json_object, "plugin") || !json_object_has_member (json_object, "name")) {
         g_set_error (error, UFO_TASK_GRAPH_ERROR, UFO_TASK_GRAPH_ERROR_JSON_KEY,
                      "Node does not have `plugin' or `name' key");
         return NULL;
@@ -787,58 +887,14 @@ create_node_from_json (JsonNode *json_node,
     task_name = json_object_get_string_member (json_object, "name");
     ufo_task_node_set_identifier (ret_node, task_name);
 
-    GHashTable *nodes_to_process = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
-
     if (json_object_has_member (json_object, "properties")) {
-        JsonObject *prop_object = json_object_get_object_member (json_object, "properties");
-        json_object_foreach_member (prop_object, add_node_to_table, nodes_to_process);
+        JsonObject *prop_object;
+        JsonSetPropertyInfo info = { .task = ret_node, .manager = manager, .error = error };
+
+        prop_object = json_object_get_object_member (json_object, "properties");
+        json_object_foreach_member (prop_object, (JsonObjectForeach) set_property_from_json, &info);
     }
 
-    if (json_object_has_member (json_object, "prop-refs")) {
-        JsonArray *prop_refs = json_object_get_array_member (json_object, "prop-refs");
-
-        for (guint i = 0; i < json_array_get_length (prop_refs); i++) {
-            const gchar *ref_name = json_array_get_string_element (prop_refs, i);
-            JsonObject *prop_set = g_hash_table_lookup (prop_sets, ref_name);
-
-            if (prop_set == NULL) {
-                g_warning ("No property set `%s' found in `prop-sets'", ref_name);
-            }
-            else {
-                json_object_foreach_member (prop_set, add_node_to_table, nodes_to_process);
-            }
-        }
-    }
-
-    g_hash_table_iter_init (&iter, nodes_to_process);
-
-    while (g_hash_table_iter_next (&iter, &key, &value)) {
-        JsonNode *node = (JsonNode *) value;
-
-        if (JSON_NODE_HOLDS_VALUE (node)) {
-            GValue val = {0,};
-            json_node_get_value (node, &val);
-            g_object_set_property (G_OBJECT(ret_node), key, &val);
-            g_value_unset (&val);
-        }
-        else if (JSON_NODE_HOLDS_OBJECT (node)) {
-            JsonObject *node_object = json_node_get_object (node);
-
-            if (json_object_has_member (node_object, "plugin")) {
-                UfoTaskNode *inner_node = create_node_from_json (node, manager, prop_sets, error);
-                g_object_force_floating (G_OBJECT (inner_node));
-                g_object_set (G_OBJECT (ret_node), key, inner_node, NULL);
-            }
-            else {
-                ufo_task_set_json_object_property (UFO_TASK (node), key, node_object);
-            }
-        }
-        else {
-            g_warning ("`%s' is neither a primitive value nor an object!", (char*)key);
-        }
-    }
-
-    g_hash_table_unref (nodes_to_process);
     return ret_node;
 }
 
@@ -906,28 +962,6 @@ handle_json_task_edge (JsonArray *array,
         g_warning ("%s", error->message);
 }
 
-static void
-handle_json_prop_set (JsonObject *object,
-                      const gchar *name,
-                      JsonNode *node,
-                      gpointer user)
-{
-    UfoTaskGraphPrivate *priv;
-    JsonObject *properties;
-
-    priv = (UfoTaskGraphPrivate *) user;
-    properties = json_object_get_object_member (object, name);
-    json_object_ref (properties);
-    g_hash_table_insert (priv->prop_sets, g_strdup (name), properties);
-}
-
-static void
-add_node_to_table (JsonObject *object, const gchar *name, JsonNode *node, gpointer table)
-{
-    GHashTable *hash_table = (GHashTable*)table;
-    g_hash_table_insert(hash_table, g_strdup(name), node);
-}
-
 static JsonObject *
 create_full_json_from_task_node (UfoTaskNode *task_node)
 {
@@ -1045,7 +1079,6 @@ ufo_task_graph_finalize (GObject *object)
     priv = UFO_TASK_GRAPH_GET_PRIVATE (object);
 
     g_hash_table_destroy (priv->json_nodes);
-    g_hash_table_destroy (priv->prop_sets);
 
     G_OBJECT_CLASS (ufo_task_graph_parent_class)->finalize (object);
 }
@@ -1070,12 +1103,7 @@ ufo_task_graph_init (UfoTaskGraph *self)
 
     priv->manager = NULL;
     priv->remote_tasks = NULL;
-
-    priv->json_nodes = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                              g_free, NULL);
-
-    priv->prop_sets = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                             g_free, (GDestroyNotify) json_object_unref);
+    priv->json_nodes = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
     priv->index = 0;
     priv->total = 1;
 }
